import asyncio
import json
from contextlib import AsyncExitStack
from dataclasses import asdict
from typing import Optional

from dotenv import load_dotenv
from mcp import ClientSession, StdioServerParameters
from mcp.client.stdio import stdio_client
from openai import AsyncOpenAI
from openai.types.chat import (
    ChatCompletionAssistantMessageParam,
    ChatCompletionMessageParam,
    ChatCompletionMessageToolCallParam,
    ChatCompletionToolMessageParam,
    ChatCompletionToolParam,
)
from openai.types.chat.chat_completion_message_tool_call_param import Function
from openai.types.shared_params.function_definition import FunctionDefinition

from .config import LLMClientConfig, LLMRequestConfig, MCPClientConfig

load_dotenv()


class MCPClient:
    def __init__(
        self,
        mpc_client_config: MCPClientConfig = MCPClientConfig(),
        llm_client_config: LLMClientConfig = LLMClientConfig(),
        llm_request_config: LLMRequestConfig = LLMRequestConfig("gpt-4o"),
    ):
        self.mpc_client_config = mpc_client_config
        self.llm_client_config = llm_client_config
        self.llm_request_config = llm_request_config
        self.llm_client = AsyncOpenAI(**asdict(self.llm_client_config))
        self.session: Optional[ClientSession] = None
        self.exit_stack = AsyncExitStack()
        self.current_server = None

        print("CLIENT CREATED")

    async def connect_to_server(self, server_name: str):
        """Connect to an MCP server using its configuration name"""

        if server_name not in self.mpc_client_config.mcpServers:
            raise ValueError(
                f"Server '{server_name}' not found in MCP client configuration"
            )

        mcp_server_config = self.mpc_client_config.mcpServers[server_name]
        if not mcp_server_config.enabled:
            raise ValueError(f"Server '{server_name}' is disabled")

        stdio_server_params = StdioServerParameters(**asdict(mcp_server_config))

        stdio_transport = await self.exit_stack.enter_async_context(
            stdio_client(stdio_server_params)
        )
        self.stdio, self.write = stdio_transport
        self.session = await self.exit_stack.enter_async_context(
            ClientSession(self.stdio, self.write)
        )

        await self.session.initialize()  # type: ignore
        self.current_server = server_name

        # List available tools
        response = await self.session.list_tools()  # type: ignore
        print(f"CLIENT CONNECT to {server_name}")
        print("AVAILABLE TOOLS", [tool.name for tool in response.tools])

    async def process_tool_call(self, tool_call) -> ChatCompletionToolMessageParam:
        match tool_call.type:
            case "function":
                tool_name = tool_call.function.name
                tool_args = json.loads(tool_call.function.arguments)
                call_tool_result = await self.session.call_tool(tool_name, tool_args)  # type: ignore

                if call_tool_result.isError:
                    raise ValueError("An error occurred while calling the tool.")

                results = []
                for result in call_tool_result.content:
                    match result.type:
                        case "text":
                            results.append(result.text)
                        case "image":
                            raise NotImplementedError("Image content is not supported")
                        case "resource":
                            raise NotImplementedError(
                                "Embedded resource is not supported"
                            )
                        case _:
                            raise ValueError(f"Unknown content type: {result.type}")

                return ChatCompletionToolMessageParam(
                    role="tool",
                    content=json.dumps({**tool_args, tool_name: results}),
                    tool_call_id=tool_call.id,
                )

            case _:
                raise ValueError(f"Unknown tool call type: {tool_call.type}")

    async def process_messages(
        self,
        messages: list[ChatCompletionMessageParam],
        llm_request_config: LLMRequestConfig | None = None,
    ) -> list[ChatCompletionMessageParam]:
        # Set up tools and LLM request config
        if not self.session:
            raise RuntimeError("Not connected to any server")
        tools = [
            ChatCompletionToolParam(
                type="function",
                function=FunctionDefinition(
                    name=tool.name,
                    description=tool.description if tool.description else "",
                    parameters=tool.inputSchema,
                ),
            )
            for tool in (await self.session.list_tools()).tools
        ]
        llm_request_config = LLMRequestConfig(
            **{
                **asdict(self.llm_request_config),
                **(asdict(llm_request_config) if llm_request_config else {}),
            }
        )

        last_message_role = messages[-1]["role"]

        match last_message_role:
            case "user":
                response = await self.llm_client.chat.completions.create(
                    messages=messages,
                    tools=tools,
                    tool_choice="auto",
                    **asdict(llm_request_config),
                )
                finish_reason = response.choices[0].finish_reason

                match finish_reason:
                    case "stop":
                        messages.append(
                            ChatCompletionAssistantMessageParam(
                                role="assistant",
                                content=response.choices[0].message.content,
                            )
                        )
                        return messages

                    case "tool_calls":
                        tool_calls = response.choices[0].message.tool_calls
                        assert tool_calls is not None
                        messages.append(
                            ChatCompletionAssistantMessageParam(
                                role="assistant",
                                tool_calls=[
                                    ChatCompletionMessageToolCallParam(
                                        id=tool_call.id,
                                        function=Function(
                                            arguments=tool_call.function.arguments,
                                            name=tool_call.function.name,
                                        ),
                                        type=tool_call.type,
                                    )
                                    for tool_call in tool_calls
                                ],
                            )
                        )
                        tasks = [
                            asyncio.create_task(self.process_tool_call(tool_call))
                            for tool_call in tool_calls
                        ]
                        messages.extend(await asyncio.gather(*tasks))
                        return await self.process_messages(messages, llm_request_config)
                    case "length":
                        raise ValueError("Length limit reached")
                    case "content_filter":
                        raise ValueError("Content filter triggered")
                    case "function_call":
                        raise NotImplementedError("Function call not implemented")
                    case _:
                        raise ValueError(f"Unknown finish reason: {finish_reason}")

            case "assistant":
                # NOTE: the only purpose of this case is to trigger other tool
                # calls based on the results of the previous tool calls
                response = await self.llm_client.chat.completions.create(
                    messages=messages,
                    tools=tools,
                    tool_choice="auto",
                    **asdict(llm_request_config),
                )
                finish_reason = response.choices[0].finish_reason

                match finish_reason:
                    case "stop":
                        # NOTE: we do not add the last response message
                        return messages

                    case "tool_calls":
                        tool_calls = response.choices[0].message.tool_calls
                        assert tool_calls is not None
                        messages.append(
                            ChatCompletionAssistantMessageParam(
                                role="assistant",
                                tool_calls=[
                                    ChatCompletionMessageToolCallParam(
                                        id=tool_call.id,
                                        function=Function(
                                            arguments=tool_call.function.arguments,
                                            name=tool_call.function.name,
                                        ),
                                        type=tool_call.type,
                                    )
                                    for tool_call in tool_calls
                                ],
                            )
                        )
                        results_messages = [
                            await self.process_tool_call(tool_call)
                            for tool_call in tool_calls
                        ]
                        messages.extend(results_messages)
                        return await self.process_messages(messages, llm_request_config)
                    case "length":
                        raise ValueError("Length limit reached")
                    case "content_filter":
                        raise ValueError("Content filter triggered")
                    case "function_call":
                        raise NotImplementedError("Function call not implemented")
                    case _:
                        raise ValueError(f"Unknown finish reason: {finish_reason}")

            case "tool":
                response = await self.llm_client.chat.completions.create(
                    messages=messages,
                    **asdict(llm_request_config),
                )
                finish_reason = response.choices[0].finish_reason

                match finish_reason:
                    case "stop":
                        messages.append(
                            ChatCompletionAssistantMessageParam(
                                role="assistant",
                                content=response.choices[0].message.content,
                            )
                        )

                        return await self.process_messages(messages, llm_request_config)
                    case "tool_calls":
                        raise ValueError(
                            "The message following a tool message cannot be a tool call"
                        )
                    case "length":
                        raise ValueError("Length limit reached")
                    case "content_filter":
                        raise ValueError("Content filter triggered")
                    case "function_call":
                        raise NotImplementedError("Function call not implemented")
                    case _:
                        raise ValueError(f"Unknown finish reason: {finish_reason}")

            case "developer":
                raise NotImplementedError("Developer messaages are not supported")
            case "system":
                raise NotImplementedError("System messages are not supported")
            case "function":
                raise NotImplementedError("System messages are not supported")
            case _:
                raise ValueError(f"Invalid message role: {last_message_role}")

    async def cleanup(self):
        """Clean up resources"""
        await self.exit_stack.aclose()



from .client import MCPClient
from .config import (
    MCPClientConfig,
    MCPServerConfig,
    LLMRequestConfig,
    LLMClientConfig,
)

__all__ = [
    "MCPClient",
    "MCPClientConfig",
    "MCPServerConfig",
    "LLMRequestConfig",
    "LLMClientConfig",
]

__version__ = "0.2.0"



# /// script
# requires-python = ">=3.11"
# dependencies = [
#     "mcp",
# ]
# ///
#
# This is a simple calculator server which implements the basic mathematical
# operations: add, subtract, multiply, and divide. It is used for testing the
# implementation of the client.

import asyncio

import mcp.server.stdio
import mcp.types as types
from mcp.server import NotificationOptions, Server
from mcp.server.models import InitializationOptions

# Initialize the server
server = Server("calculator")


@server.list_tools()
async def handle_list_tools() -> list[types.Tool]:
    """List available mathematical tools."""
    return [
        types.Tool(
            name="add",
            description="Add two numbers (+)",
            inputSchema={
                "type": "object",
                "properties": {
                    "a": {
                        "type": "number",
                        "description": "First number",
                    },
                    "b": {
                        "type": "number",
                        "description": "Second number",
                    },
                },
                "required": ["a", "b"],
            },
        ),
        types.Tool(
            name="sub",
            description="Subtract second number from first number (-)",
            inputSchema={
                "type": "object",
                "properties": {
                    "a": {
                        "type": "number",
                        "description": "First number",
                    },
                    "b": {
                        "type": "number",
                        "description": "Second number",
                    },
                },
                "required": ["a", "b"],
            },
        ),
        types.Tool(
            name="mul",
            description="Multiply two numbers (*)",
            inputSchema={
                "type": "object",
                "properties": {
                    "a": {
                        "type": "number",
                        "description": "First number",
                    },
                    "b": {
                        "type": "number",
                        "description": "Second number",
                    },
                },
                "required": ["a", "b"],
            },
        ),
        types.Tool(
            name="div",
            description="Divide first number by second number (/)",
            inputSchema={
                "type": "object",
                "properties": {
                    "a": {
                        "type": "number",
                        "description": "First number (dividend)",
                    },
                    "b": {
                        "type": "number",
                        "description": "Second number (divisor)",
                    },
                },
                "required": ["a", "b"],
            },
        ),
    ]


@server.call_tool()
async def handle_call_tool(
    name: str, arguments: dict | None
) -> list[types.TextContent | types.ImageContent | types.EmbeddedResource]:
    """Handle mathematical tool execution requests."""
    if not arguments:
        raise ValueError("Missing arguments")

    try:
        a = float(arguments.get("a", 0))
        b = float(arguments.get("b", 0))
    except (TypeError, ValueError):
        return [
            types.TextContent(
                type="text", text="Invalid input. Please provide valid numbers."
            )
        ]

    result = None
    if name == "add":
        result = a + b
    elif name == "sub":
        result = a - b
    elif name == "mul":
        result = a * b
    elif name == "div":
        if b == 0:
            return [
                types.TextContent(
                    type="text", text="Error: Division by zero is not allowed"
                )
            ]
        result = a / b
    else:
        raise ValueError(f"Unknown tool: {name}")

    return [types.TextContent(type="text", text=f"Result: {result}")]


async def main():
    """Run the server using stdin/stdout streams."""
    async with mcp.server.stdio.stdio_server() as (read_stream, write_stream):
        await server.run(
            read_stream,
            write_stream,
            InitializationOptions(
                server_name="calculator",
                server_version="0.1.0",
                capabilities=server.get_capabilities(
                    notification_options=NotificationOptions(),
                    experimental_capabilities={},
                ),
            ),
        )


if __name__ == "__main__":
    asyncio.run(main())



import os
import random

import pytest
from dotenv import load_dotenv
from pathlib import Path

from mcp_openai import MCPClient
from mcp_openai.config import (
    MCPClientConfig,
    MCPServerConfig,
    LLMRequestConfig,
    LLMClientConfig,
)

load_dotenv()


@pytest.fixture
async def client():
    server_path = Path(__file__).parent / "server.py"
    server_name = "calculator"
    client = MCPClient(
        MCPClientConfig(
            mcpServers={
                server_name: MCPServerConfig(
                    command="uv",
                    args=["run", str(server_path)],
                )
            }
        ),
        LLMClientConfig(
            api_key=os.getenv("API_KEY"),
            base_url=os.environ["BASE_URL"],
        ),
        LLMRequestConfig(
            model=os.environ["MODEL_NAME"],
            # seed=42,
        ),
    )

    await client.connect_to_server(server_name)
    yield client


@pytest.mark.asyncio()
class TestServer:
    """Test suite for calculator operations"""

    async def test_addition(self, client):
        a, b = random.randint(1, 400), random.randint(1, 400)
        messages = [{"role": "user", "content": f"What is {a} + {b}?"}]
        messages = await client.process_messages(messages)
        response = messages[-1]["content"]
        assert str(a + b) in response
        assert len(messages) == 4

    async def test_subtraction(self, client):
        a, b = random.randint(1, 400), random.randint(1, 400)
        messages = [{"role": "user", "content": f"What is {a} - {b}?"}]
        messages = await client.process_messages(messages)
        response = messages[-1]["content"]
        assert str(a - b) in response
        assert len(messages) == 4

    async def test_multiplication(self, client):
        a, b = random.randint(1, 9), random.randint(1, 9)
        messages = [{"role": "user", "content": f"What is {a} × {b}?"}]
        messages = await client.process_messages(messages)
        response = messages[-1]["content"]
        assert str(a * b) in response
        assert len(messages) == 4

    async def test_division(self, client):
        a, b = random.randint(2, 400), random.randint(1, 400)
        messages = [{"role": "user", "content": f"What is {a * b} ÷ {b}?"}]
        messages = await client.process_messages(messages)
        response = messages[-1]["content"]
        assert str(a) in response
        assert len(messages) == 4

    async def test_all_operations(self, client):
        a, b = random.randint(1, 400), random.randint(1, 400)
        content = (
            f"What is {a} + {b}?",
            f"What is {a} - {b}?",
            f"What is {a} × {b}?",
            f"What is {a * b} ÷ {b}?",
        )
        messages = [{"role": "user", "content": "\n".join(content)}]
        messages = await client.process_messages(messages)
        response = messages[-1]["content"]
        assert str(a + b) in response
        assert str(a - b) in response
        assert str(a * b) in response
        assert str(a) in response
        assert len(messages) == 7

    # @pytest.mark.skip(reason="Not implemented")
    async def test_nested_operations(self, client):
        a, b, c = random.randint(40, 60), random.randint(40, 60), random.randint(2, 5)
        messages = [
            {
                "role": "system",
                "content": (
                    "Solve the expression step by step, following the order of "
                    "operations (PEMDAS). Solve one step at a time. Use **only** "
                    "one function at a time between `add`, `sub`, `mul` and `div`."
                ),
            },
            {"role": "user", "content": f"What is ({a} + {b}) × {c}?"},
        ]
        messages = await client.process_messages(messages)
        print("LEN MESSAGES", len(messages))

        from pprint import pprint

        pprint(messages)

        response = messages[-1]["content"]
        assert str((a + b) * c) in response
        assert len(messages) == 8



from dataclasses import dataclass, field
from typing import Literal

from openai import NOT_GIVEN, NotGiven
from openai.types.chat import ChatCompletionReasoningEffort


@dataclass
class MCPServerConfig:
    """Configuration for an MCP server"""

    command: str
    args: list[str]
    env: dict[str, str] | None = None
    description: str | None = None
    enabled: bool = True


@dataclass
class MCPClientConfig:
    """Configuration for the MCP client"""

    mcpServers: dict[str, MCPServerConfig] = field(default_factory=dict)


@dataclass
class LLMClientConfig:
    """Configuration for the OpenAI (compatible) client."""

    api_key: str | None = None
    organization: str | None = None
    project: str | None = None
    base_url: str | None = None


@dataclass
class LLMRequestConfig:
    """Configuration for OpenAI client.chat.completions.create.

    Notes:
        - messages are not part of LLM request config
        - tools/tool_choice are handle directly by the MCP client
    """

    model: str
    temperature: float | NotGiven | None = NOT_GIVEN
    top_p: float | NotGiven | None = NOT_GIVEN
    max_tokens: int | NotGiven | None = NOT_GIVEN
    reasoning_effort: ChatCompletionReasoningEffort | NotGiven = NOT_GIVEN
    stream: NotGiven | Literal[False] | None = NOT_GIVEN
    stop: str | list[str] | NotGiven | None = NOT_GIVEN
    seed: int | NotGiven | None = NOT_GIVEN
    presence_penalty: float | NotGiven | None = NOT_GIVEN
    frequency_penalty: float | NotGiven | None = NOT_GIVEN
    logit_bias: dict[str, int] | NotGiven | None = NOT_GIVEN
